# This code will create the real world and online networks and save them as .graphml files, 
# to be analysed with the analysis script. The degree distribution for each network is also saved as a .csv file. 

import community
import networkx as nx
import random
from random import randint
import time


""" 
Set parameter values of interest here. For example, the following settings will create an online network with 10000 nodes, 
where each node has a 10% chance of forging a connection based upon links in the real world (realworld)
and where each node had a 10% change of linking with the node whom they are not currently connected and they share the most connections (mutual)
realworld and mutual should be set between 0% - 100% 
"""

networksize = 10000
realworld = 10
mutual = 10
start_time = time.strftime("%Y%m%d-%H%M%S")


# Network class creates real world network and then forms online network

class Network(object):

    def __init__(self, numNodes):
        
        #initialize graph with NetworkX powerlaw cluster graph
        self.RealWorld=nx.powerlaw_cluster_graph(numNodes, 3, .5, None) 
        
        #calculate the degree and betweenness for all nodes and store in a list
        self.Degree = nx.degree_centrality(self.RealWorld)
        self.Betweenness = nx.betweenness_centrality(self.RealWorld)
        
        # save the original graph density for use later
        self.OriginalDensity = nx.density(self.RealWorld)  
        
        self.fileObj=open("Degree_Report_" + "R" + str(realworld) + "M" + str(mutual)+ "_"  + start_time + ".csv", 'w')  # 'w' open for writing mode
        self.lineList2 = []
        current = str("R" + str(realworld) + "M" + str(mutual))
        self.lineList2.append(current + ", NodeID, Degree  \n")
         
        #Prints degree for the realworld network first      
        #then stores the degree and betweenness for each node
        #note that calculating betweenness here greatly increases model time, so it is commented out here
        for x in range(len(self.RealWorld.node)):
            self.RealWorld.node[x]['degree'] = self.RealWorld.degree(x)
            #self.RealWorld.node[x]['betweenness'] = self.Betweenness[x]
            print(self.RealWorld.node[x]['degree'])
            self.oneLine = " %0.6f, %0.6f, %0.6f\n" %(x, self.RealWorld.degree(x), self.RealWorld.degree(x))
            self.lineList2.append(self.oneLine)

        #save the graph for visualization in gephi
        #have to do this before storing any lists in the nodes, since gephi can't handle lists or dictionaries
        self.SaveGraph("RealWorld_" + "R" + str(realworld) + "M" + str(mutual)+ "_"  + start_time + ".graphml")
        
                
        #store each node's connections and non-connections to be used later to help recreate connections known in real world
        for x in range(len(self.RealWorld.node)):
            self.RealWorld.node[x]['connections'] = sorted(nx.all_neighbors(self.RealWorld, x))
            self.RealWorld.node[x]['non_connections'] = sorted(nx.non_neighbors(self.RealWorld, x))
            #print (self.RealWorld.node[x]['connections']) #use for testing
        
        print (nx.info(self.RealWorld)) #Print short summary of information about graph
        
    #saves a copy of the graph for visualization in gephi later
    def SaveGraph(self, graphFileName):
        nx.write_gexf(self.RealWorld, graphFileName)    
    
    def Step(self, probReal, probMutual):
             
        #create an exact copy of the Real World, saving all node attributes and edges, named CopyOnline
        self.CopyOnline = self.RealWorld.copy()
        #clear out the edges, but leave the node info
        self.WipeEdges()
        # run through the following loop to create the connections of the online version of the graph
        # stop making connections once newGraph reaches same density as RW
        while (self.OriginalDensity > nx.density(self.CopyOnline)): 
            self.CreateOnlineGraph(probReal) # adds links from original real world friend network
            self.SuggestMutualConnections(probMutual) #adds links using friend-of-friend recommendations      
        
        #nx.draw(self.CopyOnline, with_labels = True) # draw initial graph    
        self.WipeNodeAttributesAndSave()

    def WipeEdges(self):
        # Remove all the edges, but all the node data is saved
        self.CopyOnline.remove_edges_from(self.CopyOnline.edges()) 
        #nx.draw(self.CopyOnline, with_labels = True) # confirm that all the edges are gone    
    
    # Using information about connections from the real world, start to create online world
    def CreateOnlineGraph(self, probReal): 
        self.counter = 0    
        for x in range(len(self.RealWorld.node)):
            self.chance = random.randint(1, 100)
            #with some probability connect to someone you know already in the real world
            if self.chance <= probReal: 
                if len(self.RealWorld.node[x]['connections']) > 0: #make sure there are connections
                    self.randNodeIndex = randint(0, len(self.RealWorld.node[x]['connections']) -1) #choose random connection
                    self.CopyOnline.add_edge(x, sorted(self.RealWorld.node[x]['connections'])[self.randNodeIndex])
                    #update counter to show an edge was added
                    self.counter = self.counter + 1
                    if len (self.RealWorld.node[x]['connections']) > 1:
                        #remove that connection from the dictionary so other new connections can be added
                        del (self.RealWorld.node[x]['connections'])[self.randNodeIndex] 
        print("Real World Edges added: ", self.counter)
        
    def SuggestMutualConnections(self, probMutual):
        
        # counting how many edges are added   
        self.counter2 = 0 
        
        # update non-connections at the beginning of each iteration, so first nodes to go aren't advantaged with 
        # having more mutual connections within a round of updates
        
        for p in self.CopyOnline.node:
            self.CopyOnline.node[p]['non_connections'] = sorted(nx.non_neighbors(self.CopyOnline, p))
        
        ## calculate the node with the max mutual connections for each node
        for i in range (len(self.CopyOnline.node)): 
            
            self.chance = random.randint(1, 100)
            #with some probability connect to node with the max number of mutual connections
                        
            if self.chance <= probMutual: 
                tempNo = 0 #set temp num to first item being compared on unconnected list
                # check how many mutual connections are shared with every other node
                for y in range (len(self.CopyOnline.node[i]['non_connections'])):
                    #commented code below for testing
                    #RealWorld.node[i]['mutual connections with ' +str(y)] = sorted(nx.common_neighbors(RealWorld, i, y))
                    #RealWorld.node[i]['num conx with ' +str(y)] = len(sorted(nx.common_neighbors(RealWorld, i, y)))
                    #print("This is y value: ", y)
                    #if y !=i:
                    tempList =[] #create a list to store entries with max mutual number of connections
                    #compare who has the most shared neighbors, then store that index as tempNo
                    if len(sorted(nx.common_neighbors(self.CopyOnline, i, (self.CopyOnline.node[i]['non_connections'][y])))) > len(sorted(nx.common_neighbors(self.CopyOnline, i, (self.CopyOnline.node[i]['non_connections'][tempNo])))):
                        tempNo = y

                # make sure that the first option isn't always chosen if several nodes have the same max - instead select one at random, 
                # not just the first that shows up in the sorted list or that node will become highly connected
                # create a list of any that equal and save in a list to be randomly chosen from

                for y2 in range (len(self.CopyOnline.node[i]['non_connections'])):
                    if len(sorted(nx.common_neighbors(self.CopyOnline, i, (self.CopyOnline.node[i]['non_connections'][y2])))) == len(sorted(nx.common_neighbors(self.CopyOnline, i, (self.CopyOnline.node[i]['non_connections'][tempNo])))):
                        tempList.append((self.CopyOnline.node[i]['non_connections'][y2]))
                        #print(i, " " , tempList)
                #print(i, "final " , tempList)    
                oneOfMax = random.choice(tempList)        
                self.CopyOnline.node[i]['MaxMutualConnected'] = oneOfMax
                #print(i, "randMax " , oneOfMax) 

                self.CopyOnline.add_edge(i, (self.CopyOnline.node[i]['MaxMutualConnected'])) 
                self.counter2 = self.counter2 + 1    

        print("Mutual edges added: ", self.counter2)
    

        
    def WipeNodeAttributesAndSave(self):

        # Remove the list of connections and non-connections in order for gephi to accept format
        # graphml or gexf can't accept lists or arrays in node attributes, must be a single value
        
        self.fileObj=open("Degree_Report_" + "R" + str(realworld) + "M" + str(mutual)+ "_"  + start_time + ".csv", 'w')  # 'w' open for writing mode

        for x in range(len(self.CopyOnline.node)):
            del (self.CopyOnline.node[x]['connections'])
            del (self.CopyOnline.node[x]['non_connections'])
            
            self.oneLine = " %0.6f, %0.6f, %0.6f\n" %(x, self.CopyOnline.degree(x), self.CopyOnline.degree(x))
            self.lineList2.append(self.oneLine)

        self.fileObj.writelines(self.lineList2)

        # close the file
        self.fileObj.close()
        
        nx.write_gexf(self.CopyOnline, "OnlineWorld_" + "R" + str(realworld) + "M" + str(mutual) + "_" + start_time + ".graphml")


class Sim(object):
    def __init__(self, size):
        self.Network = Network(size) # Create a network with X number of nodes
        # network needs to be reinitialized each simulation run, or starts without edges        
    def Run(self):     
        for step in range(1): # won't work if you run more than once from here 
            self.Network.Step(realworld, mutual) #input probabilities of connecting to people known from real world & mutual connections (1-100= 1% - 100%)
        input("Simulation done! press Enter to exit the program.")
        
if __name__ == '__main__':
    aSim = Sim(networksize)  
    print ("Building Online Network...")    
    aSim.Run()




# In[ ]:



